
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include "jphp.hh"

Dataoutput::Dataoutput (size_t start, size_t reserve) : start (start)
{
  data.reserve (reserve);
}

size_t Dataoutput::append (const char *str)
{
  return append (str, str ? strlen (str) : 0, true);
}

size_t Dataoutput::append (const char *str, size_t len, bool search)
{
  if ( !str ) return NULL_RET;

  entry et;
  et.data = str;
  et.len = len;

  if ( search )
    {
      cache_t::iterator it = cache.find (et);
      if ( it != cache.end() )
          return it->pos;
    }

  et.pos = data.size() + start;
  if ( search ) cache.insert (et);

  data.insert (data.end(), str, str + len);
  if ( str[len - 1] != 0 )
      data.push_back (0);

  len++;

// WAGNING: it is possible that your system needs 8 byte alignment, check this out, if it is 64 bit system

  if ( len % 4 )
      fill_n (back_inserter(data), 4 - len % 4, 0);

  return et.pos;
}

void Dataoutput::write (ostream& os) const
{
  vector<char> out;
  out.reserve (data.size());
  out.insert (out.end(), data.begin(), data.end());
  os.write (&out[0], out.size());
}

HashTable * Dataoutput::serialize_hashtable_of_zval (HashTable *ht, ObjectCache<jphp_const>& consts)
{
  if ( !ht ) return reinterpret_cast<HashTable *>(NULL_RET);

#ifdef VALGRIND_MAKE_READABLE
  VALGRIND_MAKE_READABLE (&ht->bApplyProtection, (size_t) ht + sizeof (HashTable) - (size_t)&ht->bApplyProtection);
#endif

  HashTable t = *ht;

  vector<char> out;

  t.pInternalPointer = NULL;
  t.pListHead = NULL;
  t.pListTail = NULL;
  t.arBuckets = NULL;
  t.pDestructor = NULL;

  Bucket *buckets[ht->nTableSize];
  memset (buckets, 0, sizeof (Bucket *) * ht->nTableSize);

  out.insert (out.end(), reinterpret_cast<char *>(&t), reinterpret_cast<char *>(&t + 1));
  out.insert (out.end(), reinterpret_cast<char *>(buckets), reinterpret_cast<char *>(buckets + ht->nTableSize));

  Bucket *it = ht->pListHead;

  while (it)
    {
#ifdef VALGRIND_MAKE_READABLE
          VALGRIND_MAKE_READABLE (it, sizeof (Bucket) + it->nKeyLength - 1);
#endif

      char buf[sizeof(Bucket) + it->nKeyLength - 1];
      memcpy (buf, it, sizeof(Bucket) + it->nKeyLength - 1);
      Bucket *b = reinterpret_cast<Bucket *>(buf);

      b->pListNext = NULL;
      b->pListLast = NULL;
      b->pNext = NULL;
      b->pLast = NULL;
      b->pData = NULL;
      b->pDataPtr = serialize_zval (reinterpret_cast<zval *>(b->pDataPtr), consts);

      out.insert (out.end(), buf, buf + sizeof(Bucket) + it->nKeyLength - 1);

      it = it->pListNext;
    }

  return reinterpret_cast<HashTable *>(append(&out[0], out.size(), false)); // must be false so that internals dont screw
}

zval * Dataoutput::serialize_zval (zval *val, ObjectCache<jphp_const>& consts)
{
  jphp_const c;
  memset (&c, 0, sizeof (jphp_const));
  c.val = *val;
  zval& v = c.val;

  switch (v.type)
    {
      case IS_NULL:
#ifdef VALGRIND_MAKE_READABLE
        VALGRIND_MAKE_READABLE (&v, sizeof (zval));
#endif
      case IS_LONG:
      case IS_DOUBLE:
      case IS_BOOL:
        break;
      case IS_CONSTANT:
      case IS_STRING:
        c.hcode = zend_hash_func (v.value.str.val, v.value.str.len + 1);
        append_ref (v.value.str.val, v.value.str.len + 1);
        break;
      case IS_ARRAY:
      case IS_CONSTANT_ARRAY:
        v.value.str.len = 0; // valgrind fix
        v.value.ht = serialize_hashtable_of_zval (v.value.ht, consts);
        break;
      default:
        cerr << "Invalid zval type at Dataoutput::serialize_zval: " << (int)v.type << endl;
        abort ();
        break;
    }

  return reinterpret_cast<zval *>(consts.add (c));
}

HashTable * Datainput::recover_hashtable_of_zval (HashTable *ht, jphp_const *consts)
{
  recover (ht);
  if ( !ht ) return ht;

  char *ptr = reinterpret_cast<char *>(ht);
  ptr += sizeof (HashTable);
  ht->arBuckets = reinterpret_cast <Bucket **>(ptr);
  ptr += sizeof (Bucket *) * ht->nTableSize;

  Bucket *prev = NULL;
  for ( size_t i = 0; i < ht->nNumOfElements; i++ )
    {
      // recover a bucket :)
      Bucket *b = reinterpret_cast<Bucket *>(ptr);

      if ( prev )
          prev->pListNext = b;
      else
          ht->pListHead = b;

      b->pListLast = prev;

      size_t p = b->h & ht->nTableMask;
      b->pNext = ht->arBuckets[p];
      if ( ht->arBuckets[p] ) ht->arBuckets[p]->pLast = b;
      ht->arBuckets[p] = b;

      b->pDataPtr = recover_zval (reinterpret_cast<zval *>(b->pDataPtr), consts);
      b->pData = &b->pDataPtr;

      prev = b;

      ptr += sizeof (Bucket) + b->nKeyLength - 1;
    }

  ht->pListTail = prev;

  return ht;
}

zval * Datainput::recover_zval (zval *val, jphp_const *consts)
{
  return &consts[reinterpret_cast<int>(val)].val;
}

