//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#ifndef __data_hh__
#define __data_hh__

typedef struct _hashtable HashTable;

  template <typename HT>
  struct _oc_raw_hash {
    typedef HT * HTp;
    inline size_t operator() (const HTp& obj) const {
        size_t h = 5381;
        const char *arKey = reinterpret_cast<const char *>(obj);
        const char *arEnd = arKey + sizeof (HT);

        while (arKey < arEnd) {
          h += (h << 5);
          h ^= (size_t) *arKey++;
        }
        return h;
    }
  };

  template <typename HT>
  struct _oc_raw_equals {
    typedef HT * HTp;
    inline bool operator() (const HTp& l, const HTp& r) const {
      return memcmp (*l, *r, sizeof (HT)) == 0;
    }
  };

template<typename T, typename HF = _oc_raw_hash<T>, typename EF = _oc_raw_equals<T> >
class ObjectCache {
private:
public:
  inline ObjectCache (size_t reserve = 0)
  {
    data.reserve (reserve);
  }

  static const int NOT_FOUND = (size_t)-1;

  inline size_t add (const T& obj)
  {
    for ( size_t i = 0; i < data.size(); i++ )
        if ( memcmp (&data[i], &obj, sizeof(T)) == 0 ) return i;
    data.push_back (obj);
//    hash.insert (data[data.size()-1]);
    return data.size() - 1;
  }

/*
  inline size_t find (const T& obj) const
  {
    h_t::const_iterator it = obj.find(obj);
    if ( it != it.end() )
        return distance (data.begin(), it);
    return NOT_FOUND;
  }
  */

  friend inline ostream& operator<< (ostream& stream, ObjectCache<T>& cache)
  {
    stream.write (reinterpret_cast<char *>(&cache.data[0]), cache.data.size() * sizeof(T));
    return stream;
  }

  inline size_t size() const { return data.size(); }
  inline size_t data_size() const { return data.size() * sizeof (T); }

private:
  typedef vector<T> v_t;
  v_t data;
  typedef hash_set<T *, HF, EF> h_t;
  h_t hash;
};


class Dataoutput {
  struct entry {
    const char *data;
    size_t len;
    size_t pos;

    friend inline bool operator< (const entry& l, const entry& r)
    {
      int res = strncmp (l.data, r.data, l.len < r.len ? l.len : r.len);
      if ( res != 0 ) return res < 0;
      return l.len < r.len;
    }

    friend inline bool operator== (const entry& l, const entry& r)
    {
      return l.len == r.len && memcmp (l.data, r.data, l.len) == 0;
    }

    friend struct entry_hash_func {
      inline size_t operator() (const entry& e) const
      {
        size_t h = 5381;
        const char *arKey = e.data;
        const char *arEnd = arKey + e.len;

        while (arKey < arEnd) {
          h += (h << 5);
          h ^= (size_t) *arKey++;
        }
        return h;
      }
    };

    typedef hash_set<entry, entry_hash_func> cache_t;
  };

public:
  static const size_t NULL_RET = (size_t)-1;

  Dataoutput (size_t start = 0, size_t reserve = 0xffff);

  size_t append (const char *str, size_t len, bool search = true);
  size_t append (const char *str);

  void write (ostream& out) const;

  HashTable * serialize_hashtable_of_zval (HashTable *ht, ObjectCache<jphp_const>& consts);
  zval * serialize_zval (zval *val, ObjectCache<jphp_const>& consts);

  friend inline ostream& operator<< (ostream& stream, const Dataoutput& data)
  {
    data.write (stream);
    return stream;
  }

  inline char * append_str (const char *str)
  { return reinterpret_cast<char *>(append (str)); }

  inline void append_rstr (char* &str)
  { str = reinterpret_cast<char *>(append (str)); }

  template <typename T>
  inline T append_ptr (T str, size_t len, bool search = true)
  { return reinterpret_cast<T>(append (reinterpret_cast<char *>(str), len, search)); }

  template <typename T>
  void append_ref (T& where, size_t len, bool search = true)
  { where = append_ptr (where, len, search); }

  inline size_t size () const { return data.size(); }

private:
  size_t start;
  typedef vector<unsigned char> data_t;
  data_t data;
//  typedef hash_set<entry, entry_hash_func> cache_t;
  typedef entry::cache_t cache_t;
  cache_t cache;
};

class Datainput {
public:
  inline Datainput (char *data) : data (data) {}

  template <typename T>
  inline void recover (T& ref)
  {
    size_t add = reinterpret_cast<size_t>(ref);
    if ( add == Dataoutput::NULL_RET ) ref = static_cast<T>(NULL);
    else ref = reinterpret_cast<T>(data + add);
  }

  HashTable * recover_hashtable_of_zval (HashTable *ht, jphp_const *consts);
  zval * recover_zval (zval *val, jphp_const *consts);

private:
  char *data;
};

#endif // __data_hh__

