
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include <cstdio>
#include <cstdlib>

#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <cerrno>
#include <cstring>

#include <set>
#include <map>
#include <list>
#include <string>
#include <vector>
#include <deque>
#include <fstream>

#include "jphp.hh"
#include "freplace.hh"
#include "data.hh"
#include "utils.hh"
#include "jphp.hh"

typedef list<pair<vstring, zend_op> > bind_list_t;
static bind_list_t bind_functions;
static bind_list_t bind_classes;

static void my_zend_do_early_binding(TSRMLS_D)
{
  zend_op *opline = &CG(active_op_array)->opcodes[CG(active_op_array)->last-1];

  switch (opline->extended_value) {
    case ZEND_DECLARE_FUNCTION: {
        bind_functions.push_back (bind_list_t::value_type(vstring (opline->op1.u.constant.value.str.val, opline->op1.u.constant.value.str.len), *opline));
        break;
    }
    case ZEND_DECLARE_INHERITED_CLASS:
    case ZEND_DECLARE_CLASS: {
        bind_classes.push_back (bind_list_t::value_type(vstring (opline->op1.u.constant.value.str.val, opline->op1.u.constant.value.str.len), *opline));
        break;
    }
    default: {
        zend_error(E_COMPILE_ERROR, "Invalid binding type");
        return;
    }
  }

  opline->opcode = ZEND_NOP;
}

/* We have to use this because the generated data structures do not provide a clear
 * picture of the inherited class name, we only get it if class is declared in global scope,
 * but if not the inherited class name is somehwer in the op-arrays. That's icky :( */
static void scan_for_inherited_classes (const zend_op_array& op_array, hash_map<vstring, vstring, vstring_hash_func>& cmap)
{
  for ( size_t i = 0; i < op_array.size; i++ )
    {
      zend_op& op = op_array.opcodes[i];

      if ( op.opcode == ZEND_DECLARE_FUNCTION_OR_CLASS && op.extended_value == ZEND_DECLARE_INHERITED_CLASS )
        {
          char *str = op.op2.u.constant.value.str.val;
          char *sep = strchr (str, ':');

          vstring key (op.op1.u.constant.value.str.val, op.op1.u.constant.value.str.len);
          vstring val (str, sep - str);
          cmap[key] = val;
        }
    }
}

int jphp_encode (zend_file_handle *file_handle, ostream& out, bool print_header)
{
  HashTable *oft = CG(function_table);
  HashTable *oct = CG(class_table);
  HashTable nft;
  HashTable nct;
  zend_hash_init_ex(&nft, 127, NULL, ZEND_FUNCTION_DTOR, 1, 0);
  zend_hash_init_ex(&nct, 127, NULL, ZEND_CLASS_DTOR, 1, 0);
  CG(function_table) = &nft;
  CG(class_table) = &nct;

  bind_functions.clear ();
  bind_classes.clear ();

  freplace_data r;
  freplace (&r, zend_do_early_binding, my_zend_do_early_binding);

  zend_op_array *op = compile_file(file_handle, 0);

  frestore (&r);

  bind_list_t bind_f (bind_functions);
  bind_list_t bind_c (bind_classes);

  CG(function_table) = oft;
  CG(class_table) = oct;

  string opened_path = file_handle->opened_path ? file_handle->opened_path : file_handle->filename;

  if ( !op )
    {
      zend_hash_destroy (&nft);
      zend_hash_destroy (&nct);
      return FAILURE;
    }

  /* Compilation is done, now move on to writing headers and encoding oparrays. */

  // at first handle all the registered functions for this file
  Dataoutput data;
  ObjectCache<jphp_const> consts;
  hash_map<vstring, vstring, vstring_hash_func> class_extends;

  if ( print_header ) out << "#!/usr/local/bin/php\n<?php */%%JPHP|x86|32|001;\n";

  php_hash<zend_function> pht(&nft);
  php_hash<zend_class_entry> phc(&nct);

  vector<jphp_fentry> fentrys;
  fentrys.reserve (pht.size() + 1);

  map<vstring, int> f2index;

  jphp_fentry entry;
  memset (&entry, 0, sizeof (jphp_fentry));
  entry.op_array = op;
  entry.bind_at_load = 0;
  entry.name = data.append_str (NULL);
  entry.name_len = 0;
  entry.name_hcode = 0;
  fentrys.push_back (entry);

  scan_for_inherited_classes (*op, class_extends);

  int i = 1;
  for ( bind_list_t::iterator it = bind_f.begin(); it != bind_f.end(); it++, i++ )
    {
      zend_function *f = pht.find (it->first.c_str(), it->first.size());

      jphp_fentry entry;
      memset (&entry, 0, sizeof (jphp_fentry));
      entry.op_array = &f->op_array;
      entry.bind_at_load = 1;
      entry.name = data.append_str (f->common.function_name);
      entry.name_len = strlen (f->common.function_name);
      entry.name_hcode = zend_hash_func (f->common.function_name, entry.name_len + 1);

      scan_for_inherited_classes (f->op_array, class_extends);

      fentrys.push_back (entry);
      f2index[it->first] = i;
    }

  pht.reset();
  while ( zend_function *f = pht.get_data () )
    {
      vstring key = pht.get_string_key ();
      pht.forw ();

      if ( f2index.find(key) != f2index.end() ) { continue; }

      jphp_fentry entry;
      memset (&entry, 0, sizeof (jphp_fentry));
      entry.op_array = &f->op_array;
      entry.bind_at_load = 0;
      entry.name = data.append_str (f->common.function_name);
      entry.name_len = strlen (f->common.function_name);
      entry.name_hcode = zend_hash_func (f->common.function_name, entry.name_len + 1);

      scan_for_inherited_classes (f->op_array, class_extends);

      fentrys.push_back (entry);
      f2index[key] = i;

      i++;
    }

  // Now working on all the class entrys
  vector<jphp_centry> centrys;
  centrys.reserve (phc.size() + 1);
  map<vstring, int> c2index;

  i = 0;
  for ( bind_list_t::iterator it = bind_c.begin(); it != bind_c.end(); it++ )
    {
      jphp_centry entry;
      memset (&entry, 0, sizeof (jphp_centry));
      entry.centry = phc.find (it->first.c_str(), it->first.size());
      entry.bind_at_load = 1;

      if ( it->second.extended_value == ZEND_DECLARE_INHERITED_CLASS )
        {
          char *sep = strchr (it->second.op2.u.constant.value.str.val, ':');
          vstring str(it->second.op2.u.constant.value.str.val, sep - it->second.op2.u.constant.value.str.val);
          entry.parent = const_cast<char *>(str.c_str()); // UUUGLY
          entry.parent_len = sep - it->second.op2.u.constant.value.str.val;
          entry.parent_hcode = zend_hash_func (entry.parent, entry.parent_len + 1);
          data.append_ref (entry.parent, entry.parent_len + 1);
        }
      else
        {
          entry.parent = NULL;
          entry.parent_len = entry.parent_hcode = 0;
          data.append_ref (entry.parent, 0);
        }

      php_hash<zend_function> ht(&entry.centry->function_table);
      ht.reset ();
      while ( zend_function *f = ht.each() )
          scan_for_inherited_classes (f->op_array, class_extends);

      centrys.push_back (entry);
      c2index[it->first] = i++;
    }

  phc.reset();
  while ( zend_class_entry *c = phc.each () )
    {
      php_hash<zend_function> ht(&c->function_table);
      while ( zend_function *f = ht.each() )
          scan_for_inherited_classes (f->op_array, class_extends);
    }

  phc.reset();
  while ( zend_class_entry *c = phc.get_data () )
    {
      vstring key = phc.get_string_key ();
      phc.forw ();

      if ( c2index.find(key) != c2index.end() ) { continue; }

      jphp_centry entry;
      memset (&entry, 0, sizeof (jphp_centry));

      hash_map<vstring,vstring,vstring_hash_func>::iterator it = class_extends.find(key);
      if ( it != class_extends.end() )
        {
          entry.parent = data.append_str(it->second.c_str());
          entry.parent_len = strlen (it->second.c_str());
          entry.parent_hcode = zend_hash_func (const_cast<char *>(it->second.c_str()), entry.parent_len + 1);
        }
      else
        {
          entry.parent = NULL;
          entry.parent_len = entry.parent_hcode = 0;
          data.append_ref (entry.parent, 0);
        }

      entry.centry = c;
      entry.bind_at_load = 0;

      centrys.push_back (entry);

      f2index[key] = i++;
    }

  // Now outputing the header, fentrys, f-op-arrays, centrys, c-op-arrays
  jphp_header head;
  memset (&head, 0, sizeof (jphp_header));
  head.path = data.append_str (opened_path.c_str());
  head.function_count = fentrys.size();
  head.class_count = centrys.size();
  out.write (reinterpret_cast<char *>(&head), sizeof (jphp_header));

  out.write (reinterpret_cast<char *>(&fentrys[0]), fentrys.size() * sizeof (jphp_fentry));

  for ( size_t i = 0; i < fentrys.size(); i++ )
    {
      jphp_fentry &entry = fentrys[i];
      jphp_optimize (*entry.op_array, f2index, data, consts, out);
    }

  for ( size_t i = 0; i < centrys.size(); i++ )
    {
      jphp_centry &entry = centrys[i];
      zend_class_entry *c = entry.centry;

      entry.name = data.append_str (c->name);
      entry.name_len = c->name_length;
      entry.name_hcode = zend_hash_func (c->name, entry.name_len + 1);

      zend_class_entry e = *c;
      entry.centry = &e;
      data.append_ref (e.name, e.name_length + 1);
      e.parent = NULL;
      e.refcount = NULL;
      e.builtin_functions = NULL;
      e.handle_function_call = NULL;
      e.handle_property_get = NULL;
      e.handle_property_set = NULL;
      memset (&e.default_properties, 0, sizeof (HashTable));
      memset (&e.function_table, 0, sizeof (HashTable));
      entry.default_properties = data.serialize_hashtable_of_zval (&c->default_properties, consts);
      data.append_ref (entry.centry, sizeof (zend_class_entry), false); // must be false
      entry.function_count = c->function_table.nNumOfElements;
      entry.function_table = &c->function_table;
    }

  out.write (reinterpret_cast<char *>(&centrys[0]), centrys.size() * sizeof (jphp_centry));

  for ( size_t i = 0; i < centrys.size(); i++ )
    {
      jphp_centry &entry = centrys[i];
      HashTable& ft = *entry.function_table;
      ft.pInternalPointer = NULL;
      Bucket *b = ft.pListHead;

      out.write (reinterpret_cast<char *>(&ft), sizeof (HashTable));

      Bucket *bar[ft.nTableSize];
      memset (bar, 0, sizeof (Bucket *) * ft.nTableSize);
      out.write (reinterpret_cast<char *>(bar), sizeof (Bucket *) * ft.nTableSize);

      while (b)
        {
#ifdef VALGRIND_MAKE_READABLE
          VALGRIND_MAKE_READABLE (b, sizeof (Bucket) + b->nKeyLength - 1);
#endif

          char buf[sizeof (Bucket) + b->nKeyLength - 1];
          memcpy (buf, b, sizeof (Bucket) + b->nKeyLength - 1);
          Bucket *tmpb = reinterpret_cast<Bucket *>(buf);
          tmpb->pData = tmpb->pDataPtr = tmpb->pListNext = tmpb->pListLast = tmpb->pNext = tmpb->pLast = NULL;

          out.write (reinterpret_cast<char *>(tmpb), sizeof (Bucket) + b->nKeyLength - 1);

          zend_op_array *a = reinterpret_cast<zend_op_array *>(b->pData);
          jphp_optimize (*a, f2index, data, consts, out);
          b = b->pListNext;
        }
    }

  size_t cs = consts.size();
  out.write (reinterpret_cast<char *>(&cs), sizeof(size_t));

  out << consts;
  out << data;

  destroy_op_array (op);
  zend_hash_destroy (&nft);
  zend_hash_destroy (&nct);

  return SUCCESS;
}

int jphp_encode (const char *srcpath, const char *targetpath, bool print_header)
{
  zend_file_handle file_handle;
  file_handle.filename = const_cast<char *>(srcpath);
  file_handle.free_filename = 0;
  file_handle.type = ZEND_HANDLE_FILENAME;
  file_handle.opened_path = NULL;

  ofstream out(targetpath, ios::binary);

  if ( !out.is_open () )
    {
      zend_error (E_WARNING, "Unable to open given target path [%s] for encode: %s", targetpath, strerror (errno));
      return FAILURE;
    }

  int ret = jphp_encode (&file_handle, out, print_header );

  if ( ret == FAILURE )
      unlink (targetpath);

  out.close ();

  zend_destroy_file_handle (&file_handle TSRMLS_CC);

  return ret;
}

