
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

// This code is partly taken from the OSSP mm library,
// written by Ralf S. Engelschall <rse@engelschall.com>,
// at http://www.ossp.org/.

#include <cstdio>
#include <cstdlib>
#include <cerrno>
#include <cstring>

#include "mm.hh"

template<typename T>
static inline T min_of (const T& a, const T& b) { return a < b ? a : b; }

template<typename T>
static inline T max_of (const T& a, const T& b) { return a > b ? a : b; }

template<typename T>
static inline T abs_of (const T& a) { return ((a) < 0 ? -(a) : (a)); }

union mem_word {
    void  *mw_vp;
    void (*mw_fp)(void);
    char  *mw_cp;
    long   mw_l;
    double mw_d;
};
typedef union mem_word mem_word;
#define SIZEOF_mem_word (sizeof(mem_word))

union mem_chunk_mc_u {
    struct mem_chunk *mc_next; /* really used when it's free         */
    mem_word          mc_base; /* virtually used when it's allocated */
};
struct mem_chunk {
    size_t mc_size;      /* physical size   */
    size_t mc_usize;     /* user known size */
    union mem_chunk_mc_u mc_u;
};
typedef struct mem_chunk mem_chunk;
#define SIZEOF_mem_chunk (sizeof(mem_chunk)-sizeof(union mem_chunk_mc_u))

struct mem_pool {
   size_t     mp_size;
   size_t     mp_offset;
   mem_chunk  mp_freechunks;
   mem_word   mp_base;
};
typedef struct mem_pool mem_pool;
#define SIZEOF_mem_pool (sizeof(mem_pool)-SIZEOF_mem_word)

typedef mem_pool MM;

static inline size_t mm_core_align2word(size_t size)
{
    return ((1+((size-1) / SIZEOF_mem_word)) * SIZEOF_mem_word);
}

/*
 * Insert a chunk to the list of free chunks. Algorithm used is:
 * Insert in sorted manner to the list and merge with previous
 * and/or next chunk when possible to form larger chunks out of
 * smaller ones.
 */
static void mm_insert_chunk(MM *mm, mem_chunk *mcInsert)
{
    mem_chunk *mc;
    mem_chunk *mcPrev;
    mem_chunk *mcNext;

    mc = &(mm->mp_freechunks);
    while (mc->mc_u.mc_next != NULL && (char *)(mc->mc_u.mc_next) < (char *)mcInsert)
        mc = mc->mc_u.mc_next;
    mcPrev = mc;
    mcNext = mc->mc_u.mc_next;
    if (mcPrev == mcInsert || mcNext == mcInsert) {
        return;
    }
    if ((char *)mcPrev+(mcPrev->mc_size) == (char *)mcInsert &&
        (mcNext != NULL && (char *)mcInsert+(mcInsert->mc_size) == (char *)mcNext)) {
        /* merge with previous and next chunk */
        mcPrev->mc_size += mcInsert->mc_size + mcNext->mc_size;
        mcPrev->mc_u.mc_next = mcNext->mc_u.mc_next;
        mm->mp_freechunks.mc_usize -= 1;
    }
    else if ((char *)mcPrev+(mcPrev->mc_size) == (char *)mcInsert) {
        /* merge with previous chunk */
        mcPrev->mc_size += mcInsert->mc_size;
    }
    else if (mcNext != NULL && (char *)mcInsert+(mcInsert->mc_size) == (char *)mcNext) {
        /* merge with next chunk */
        mcInsert->mc_size += mcNext->mc_size;
        mcInsert->mc_u.mc_next = mcNext->mc_u.mc_next;
        mcPrev->mc_u.mc_next = mcInsert;
    }
    else {
        /* no merging possible, so insert as new chunk */
        mcInsert->mc_u.mc_next = mcNext;
        mcPrev->mc_u.mc_next = mcInsert;
        mm->mp_freechunks.mc_usize += 1;
    }
    return;
}

/*
 * Retrieve a chunk from the list of free chunks.  Algorithm used
 * is: Search for minimal-sized chunk which is larger or equal
 * than the request size. But when the retrieved chunk is still a
 * lot larger than the requested size, split out the requested
 * size to not waste memory.
 */
static mem_chunk *mm_retrieve_chunk(MM *mm, size_t size)
{
    mem_chunk *mc;
    mem_chunk **pmcMin;
    mem_chunk *mcRes;
    size_t sMin;
    size_t s;

    if (size == 0)
        return NULL;
    if (mm->mp_freechunks.mc_usize == 0)
        return NULL;

    /* find best-fitting chunk */
    pmcMin = NULL;
    sMin = mm->mp_size; /* initialize with maximum possible */
    mc = &(mm->mp_freechunks);
    while (mc->mc_u.mc_next != NULL) {
        s = mc->mc_u.mc_next->mc_size;
        if (s >= size && s < sMin) {
            pmcMin = &(mc->mc_u.mc_next);
            sMin = s;
            if (s == size)
                break;
        }
        mc = mc->mc_u.mc_next;
    }

    /* create result chunk */
    if (pmcMin == NULL)
        mcRes = NULL;
    else {
        mcRes = *pmcMin;
        if (mcRes->mc_size >= (size + min_of(2*size,128u))) {
            /* split out in part */
            s = mcRes->mc_size - size;
            mcRes->mc_size = size;
            /* add back remaining chunk part as new chunk */
            mc = (mem_chunk *)((char *)mcRes + size);
            mc->mc_size = s;
            mc->mc_u.mc_next = mcRes->mc_u.mc_next;
            *pmcMin = mc;
        }
        else {
            /* split out as a whole */
            *pmcMin = mcRes->mc_u.mc_next;
            mm->mp_freechunks.mc_usize--;
        }
    }

    return mcRes;
}


MemMan::MemMan (void *ptr, size_t usize)
{
  /* fill in the memory pool structure */
  mm = (MM *)ptr;
  mm->mp_size    = usize;
  mm->mp_offset  = SIZEOF_mem_pool;

  /* first element of list of free chunks counts existing chunks */
  mm->mp_freechunks.mc_size      = 0; /* has to be 0 forever */
  mm->mp_freechunks.mc_usize     = 0; /* counts chunks */
  mm->mp_freechunks.mc_u.mc_next = NULL;
}

void * MemMan::alloc (size_t usize)
{
  mem_chunk *mc;
  size_t size;
  void *vp;

  if (mm == NULL || usize == 0)
      return NULL;
  size = mm_core_align2word(SIZEOF_mem_chunk+usize);
  if ((mc = mm_retrieve_chunk(mm, size)) != NULL) {
      mc->mc_usize = usize;
      return &(mc->mc_u.mc_base.mw_cp);
  }
  if ((mm->mp_size - mm->mp_offset) < size) {
      errno = ENOMEM;
      return NULL;
  }
  mc = (mem_chunk *)((char *)mm + mm->mp_offset);
  mc->mc_size  = size;
  mc->mc_usize = usize;
  vp = (void *)&(mc->mc_u.mc_base.mw_cp);
  mm->mp_offset += size;
  return vp;
}

void MemMan::free (void *ptr)
{
  mem_chunk *mc;

  if (mm == NULL || ptr == NULL)
      return;
  mc = (mem_chunk *)((char *)ptr - SIZEOF_mem_chunk);
  mm_insert_chunk(mm, mc);
  return;
}

void * MemMan::realloc (void *ptr, size_t usize)
{
  size_t size;
  mem_chunk *mc;
  void *vp;

  if (mm == NULL || usize == 0)
      return NULL;
  if (ptr == NULL)
      return alloc(usize); /* POSIX.1 semantics */
  mc = (mem_chunk *)((char *)ptr - SIZEOF_mem_chunk);
  if (usize <= mc->mc_usize) {
      mc->mc_usize = usize;
      return ptr;
  }
  size = mm_core_align2word(SIZEOF_mem_chunk+usize);
  if (size <= mc->mc_size) {
      mc->mc_usize = usize;
      return ptr;
  }
  if ((vp = alloc(usize)) == NULL)
      return NULL;
  memcpy(vp, ptr, usize);
  free(ptr);
  return vp;
}

void * MemMan::calloc (size_t number, size_t usize)
{
  void *vp;

  if (mm == NULL || number*usize == 0)
      return NULL;
  if ((vp = alloc(number*usize)) == NULL)
      return NULL;
  memset(vp, 0, number*usize);
  return vp;
}

char * MemMan::strdup (const char *str)
{
  int n;
  void *vp;

  if (mm == NULL || str == NULL)
      return NULL;
  n = strlen(str);
  if ((vp = alloc(n+1)) == NULL)
      return NULL;
  memcpy(vp, str, n+1);
  return static_cast<char *>(vp);
}

size_t MemMan::size_of (const void *ptr) const
{
  mem_chunk *mc;

  if (mm == NULL || ptr == NULL)
      return static_cast<size_t>(-1);
  mc = (mem_chunk *)((char *)ptr - SIZEOF_mem_chunk);
  return mc->mc_usize;
}

size_t MemMan::available () const
{
  mem_chunk *mc;
  int nFree;

  nFree = mm->mp_size-mm->mp_offset;
  mc = &(mm->mp_freechunks);
  while (mc->mc_u.mc_next != NULL) {
      mc = mc->mc_u.mc_next;
      nFree += mc->mc_size;
  }
  return nFree;
}

MemMan::~MemMan ()
{
}

