
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include <cstdio>
#include <cerrno>
#include <cstring>

#include <signal.h>

#include <set>
#include <map>
#include <list>
#include <string>
#include <vector>
#include <deque>

#include "jphp.hh"
#include "optimize.hh"

namespace {
  void fix_brk_cont (Cell::brk_cont_t& arr, Cell *from, Cell *to)
  {
    for ( size_t i = 0; i < arr.size(); i++ )
      {
        if ( arr[i].brk == from ) arr[i].brk = to;
        if ( arr[i].cont == from ) arr[i].cont = to;
      }
  }
};

void jphp_optimize (const zend_op_array& op_array,
    const map<vstring, int>& f2index, Dataoutput& data, ObjectCache<jphp_const>& consts, ostream& out)
{
  if ( op_array.start_op ) return;

  vector<Cell *> cref;
  typedef list<Cell *> cells_t;
  cells_t cells;

  cref.reserve (op_array.size);

  for ( zend_uint i = 0; i < op_array.size; i++ )
    {
#ifdef VALGRIND_MAKE_READABLE
      if ( op_array.opcodes[i].result.op_type == IS_UNUSED )
          VALGRIND_MAKE_READABLE (&op_array.opcodes[i].result, sizeof (znode));
      if ( op_array.opcodes[i].op1.op_type == IS_UNUSED )
          VALGRIND_MAKE_READABLE (&op_array.opcodes[i].op1, sizeof (znode));
      if ( op_array.opcodes[i].op2.op_type == IS_UNUSED )
          VALGRIND_MAKE_READABLE (&op_array.opcodes[i].op2, sizeof (znode));
#endif

      Cell *c = new Cell (&op_array.opcodes[i]);
      cells.push_back (c);
      cref.push_back (c);
    }

  // Fill in jump points
  for ( size_t i = 0; i < cref.size(); i++ )
    {
      Cell *c = cref[i];

      switch (c->node.opcode)
        {
          case ZEND_RETURN:
            if ( c->node.opcode == ZEND_RETURN && c->op1.u.constant.type == IS_LONG )
              {
                long tmp = c->op1.u.constant.value.lval;
                memset (&c->op1.u.constant, 0, sizeof (zval));
                c->op1.u.constant.type = IS_LONG;
                c->op1.u.constant.value.lval = tmp;
                c->op1.u.constant.refcount = 1;
              }
            c->node.op1.EA_type = 0;
            continue;
            break;
          case ZEND_JMP:
            {
              Cell *cn = cref[c->op1.u.opline_num];
              c->alt_jmp_to (cn);
              c->node.jext_type = JEXT_ALT_JMP;
              c->node.op1.op_type = JOP_UNUSED;
              continue;
              break;
            }
          case ZEND_JMP_NO_CTOR:
          case ZEND_FE_FETCH:
          case ZEND_JMPZ:
          case ZEND_JMPNZ:
          case ZEND_JMPZ_EX:
          case ZEND_JMPNZ_EX:
            {
              Cell *cn = cref[c->op2.u.opline_num];
              c->alt_jmp_to (cn);
              c->node.jext_type = JEXT_ALT_JMP;
              c->node.op2.op_type = JOP_UNUSED;
              break;
            }
          case ZEND_JMPZNZ:
            {
              Cell *cn = cref[c->op2.u.opline_num];
              c->node.op2.op_type = JOP_JMP;
              c->j2_next = cn;
              cn->jump_from.insert (c);
              c->jump_to.insert (cn);

              cn = cref[c->extended_value];
              c->alt_jmp_to (cn);
              c->node.jext_type = JEXT_ALT_JMP;

              continue;
              break;
            }
          default:
            break;
        }
    }

  // Fill in the break and cont arrays, i don't really understand right now how this brk-cont stuff works
  Cell::brk_cont_t brk_cont;
  if ( op_array.brk_cont_array )
    {
      brk_cont.reserve (op_array.last_brk_cont);
      for ( size_t i = 0; i < op_array.last_brk_cont; i++ )
        {
          Cell::brk_cont_str bc;
          bc.brk = op_array.brk_cont_array[i].brk == -1 ? NULL : cref[op_array.brk_cont_array[i].brk];
          bc.cont = op_array.brk_cont_array[i].cont == -1 ? NULL : cref[op_array.brk_cont_array[i].cont];
          bc.parent = op_array.brk_cont_array[i].parent;
          brk_cont.push_back (bc);
        }
    }

  // Add first starting operand ZEND_JEXT_START, this must always be the first operand
  Cell *first = *cells.begin();
  Cell *c = new Cell ();
  c->node.opcode = ZEND_JEXT_START;
  c->node.lineno = op_array.opcodes[0].lineno;
  c->node.jext_type = JEXT_HEADER;
  c->node.jex.header = NULL; // this will be filled in by the decode
  cells.insert (cells.begin(), c);
  
  // Fix the ZEND_DECLARE_FUNCTION_OR_CLASS calls, also binary oparations
  for ( cells_t::iterator it = cells.begin(); it != cells.end(); it++ )
    {
      Cell& c = **it;
      if ( c.node.opcode == ZEND_DECLARE_FUNCTION_OR_CLASS )
        {
          vstring fn(c.op1.u.constant.value.str.val, c.op1.u.constant.value.str.len);
          zend_uint extended_value = c.node.jex.extended_value;
          assert ( f2index.find(fn) != f2index.end() );

          switch (c.node.jex.extended_value)
            {
              case ZEND_DECLARE_FUNCTION:
                {
                  c.node.jex.fentry = reinterpret_cast<jphp_fentry *>(f2index.find(fn)->second);
                  c.node.jext_type = JEXT_FENTRY;
                  break;
                }
              case ZEND_DECLARE_INHERITED_CLASS:
              case ZEND_DECLARE_CLASS:
                {
                  c.node.jex.centry = reinterpret_cast<jphp_centry *>(f2index.find(fn)->second);
                  c.node.jext_type = JEXT_CENTRY;
                  break;
                }
            }

          memset(&c.node.op1, 0, sizeof(jnode));
          c.node.op1.op_type = JOP_UNUSED;
          memset(&c.op1, 0, sizeof(znode));
          SET_UNUSED(c.op1);
          memset(&c.node.op2, 0, sizeof(jnode));
          c.node.op2.op_type = JOP_UNUSED;
          memset(&c.op2, 0, sizeof(znode));
          SET_UNUSED(c.op2);

          c.node.op1.u.register_type = extended_value;
          c.node.op1.op_type = JOP_REG_TYPE;

          continue;
        }

      switch (c.node.opcode)
        {
          case ZEND_RECV_INIT:  
            c.node.op2.EA_type = 0;
            break;
          case ZEND_INCLUDE_OR_EVAL:  // bugs in Zend
          case ZEND_ISSET_ISEMPTY:
            c.node.op2.op_type = JOP_CONST;
            c.op2.op_type = IS_CONST;
            break;

          case ZEND_ASSIGN_ADD:
          case ZEND_ASSIGN_SUB:
          case ZEND_ASSIGN_MUL:
          case ZEND_ASSIGN_DIV:
          case ZEND_ASSIGN_MOD:
          case ZEND_ASSIGN_SL:
          case ZEND_ASSIGN_SR:
          case ZEND_ASSIGN_CONCAT:
          case ZEND_ASSIGN_BW_OR:
          case ZEND_ASSIGN_BW_AND:
          case ZEND_ASSIGN_BW_XOR:
            c.node.jex.bop.op = reinterpret_cast<int (*)(zval *, zval *, zval *)>(c.node.opcode);
            c.node.opcode = ZEND_JEXT_BOP_ASSIGN;
            c.node.jext_type = JEXT_BOP;
            break;

          case ZEND_ADD:
          case ZEND_SUB:
          case ZEND_MUL:
          case ZEND_DIV:
          case ZEND_MOD:
          case ZEND_SL:
          case ZEND_SR:
          case ZEND_CONCAT:
          case ZEND_BW_OR:
          case ZEND_BW_AND:
          case ZEND_BW_XOR:
          case ZEND_IS_IDENTICAL:
          case ZEND_IS_NOT_IDENTICAL:
          case ZEND_IS_EQUAL:
          case ZEND_IS_NOT_EQUAL:
          case ZEND_IS_SMALLER:
          case ZEND_IS_SMALLER_OR_EQUAL:
          case ZEND_BOOL_XOR:
            c.node.jex.bop.op = reinterpret_cast<int (*)(zval *, zval *, zval *)>(c.node.opcode);
            c.node.opcode = ZEND_JEXT_BOP;
            c.node.jext_type = JEXT_BOP;
            break;
          case ZEND_FETCH_R: {
            if ( c.op1.op_type == IS_CONST && c.op1.u.constant.type == IS_STRING && c.op2.u.fetch_type == ZEND_FETCH_LOCAL )
                c.node.opcode = ZEND_JEXT_FETCH_LOCAL_R_BY_CONST;
            break;
          }
          case ZEND_FETCH_W: {
            if ( c.op1.op_type == IS_CONST && c.op1.u.constant.type == IS_STRING && c.op2.u.fetch_type == ZEND_FETCH_LOCAL )
                c.node.opcode = ZEND_JEXT_FETCH_LOCAL_W_BY_CONST;
            break;
          }
          case ZEND_FETCH_RW: {
            if ( c.op1.op_type == IS_CONST && c.op1.u.constant.type == IS_STRING && c.op2.u.fetch_type == ZEND_FETCH_LOCAL )
                c.node.opcode = ZEND_JEXT_FETCH_LOCAL_RW_BY_CONST;
            break;
          }
          case ZEND_FETCH_CONSTANT: {
            break;
          }
        }
    }

  /*
    Find all fetches(r). This is a horrible way of doing it.
  */
  for ( cells_t::iterator it = cells.begin(); it != cells.end(); )
    {
      Cell& c = **it;
      if ( c.node.opcode == ZEND_JEXT_FETCH_LOCAL_R_BY_CONST )
        {
          uint vid = c.node.result.u.var;
          zval const_val = c.op1.u.constant;
          bool found = false;

          for ( cells_t::iterator jt = it; jt != cells.end(); jt++ )
            {
              Cell& c2 = **jt;
              found = false;
              switch (c2.node.opcode)
                {
                  case ZEND_ASSIGN:
                  case ZEND_JEXT_BOP_ASSIGN:
                  case ZEND_JEXT_BOP:
                  case ZEND_JMPZ:
                  case ZEND_JMPNZ:
                  case ZEND_JMPZ_EX:
                  case ZEND_JMPNZ_EX:
                  case ZEND_EXIT:
                  case ZEND_PRINT:
                  case ZEND_ECHO:
                  case ZEND_ADD_CHAR:
                  case ZEND_ADD_STRING:
                  case ZEND_ADD_VAR:
                  case ZEND_BOOL:
                  case ZEND_NEW:
                  case ZEND_CAST:
                  case ZEND_INCLUDE_OR_EVAL:
                  case ZEND_UNSET_VAR:
                  case ZEND_FE_FETCH:
                  case ZEND_QM_ASSIGN:
                  case ZEND_BW_NOT:
                  case ZEND_BOOL_NOT:
                  {
                    if ( c2.node.op1.op_type == JOP_VAR )
                      {
                        if ( c2.node.op1.u.var == vid )
                          {
                            c2.node.op1.op_type = JOP_FETCH;
                            c2.op1.op_type = IS_CONST;
                            c2.op1.u.constant = const_val;
                            found = true;
                          }
                      }

                    if ( c2.node.op2.op_type == JOP_VAR )
                      {
                        if ( c2.node.op2.u.var == vid )
                          {
                            c2.node.op2.op_type = JOP_FETCH;
                            c2.op2.op_type = IS_CONST;
                            c2.op2.u.constant = const_val;
                            found = true;
                          }
                      }
                    break;
                  }
                }

              if ( found )
                { 
                  cells_t::iterator n = it;
                  n++;
                  (*it)->replace (*n);
                  fix_brk_cont (brk_cont, *it, *n);
                  delete *it;
                  it = cells.erase (it);
                  break;
                }
            }

          if ( found ) continue;
        }
      it++;
    }
  
// do the assign fetch(w) speedup hack
  for ( cells_t::iterator it = cells.begin(); it != cells.end(); it++ )
    {
      if ( it == cells.end() ) continue;
      Cell& c = **it;
      if ( c.node.opcode != ZEND_ASSIGN && c.node.opcode != ZEND_JEXT_BOP_ASSIGN ) continue;

      cells_t::iterator p = it;
      Cell *parent = *--p;

      if ( parent->node.opcode == ZEND_JEXT_FETCH_LOCAL_W_BY_CONST && parent->node.result.u.var == c.node.op1.u.var )
        {
          c.node.op1.op_type = JOP_FETCH;
          c.op1.op_type = IS_CONST;
          c.op1.u.constant = parent->op1.u.constant;
          parent->replace (&c);
          fix_brk_cont (brk_cont, parent, &c);
          delete parent;
          cells.erase (p);
        }
    }

// do some weird jump optimizations
  for ( cells_t::iterator it = cells.begin(); it != cells.end(); it++ )
    {
      Cell& c = **it;
      if ( c.next_jmp )
        {
          Cell *to = c.next_jmp;
          if ( to->node.opcode == ZEND_JMP )
            {
              to->jump_from.erase (&c);
              c.jump_to.erase (to);
              c.next_jmp = to->next_jmp;
              c.jump_to.insert (c.next_jmp);
              c.next_jmp->jump_from.insert (&c);
            }
        }

      if ( c.j2_next )
        {
          Cell *to = c.j2_next;
          if ( to->node.opcode == ZEND_JMP )
            {
              to->jump_from.erase (&c);
              c.jump_to.erase (to);
              c.j2_next = to->next_jmp;
              c.jump_to.insert (c.next_jmp);
              c.j2_next->jump_from.insert (&c);
            }
        }
    }

// Remove NOP-s
  for ( cells_t::iterator it = cells.begin(); it != cells.end(); )
    {
      Cell& c = **it;
      if ( c.node.opcode == ZEND_NOP )
        {
          it = cells.erase (it);
          fix_brk_cont (brk_cont, &c, *it);
          c.replace (*it);
          delete &c;
        }
      else
        {
          it++;
        }
    }

// If two consequal JMP-s without the second being used, remove first
  for ( cells_t::iterator it = cells.begin(); it != cells.end(); )
    {
      if ( it == cells.begin() ) { it++; continue; }
      Cell& c = **it;
      cells_t::iterator p = it;
      --p;
      Cell& parent = **p;

      if ( parent.node.opcode == ZEND_JMP && c.node.opcode == ZEND_JMP && !c.jump_from.size() )
        {
          c.next_jmp->jump_from.erase (&c);
          it = cells.erase (it);
          fix_brk_cont (brk_cont, &c, *it);
          delete &c;
          continue;
        }
      it++;
    }

  // Fill in the order id of cells 
  cref.clear ();
  cref.reserve (cells.size ());

  for ( cells_t::iterator it = cells.begin(); it != cells.end(); it++ )
    {
      (*it)->order_id = cref.size();
      cref.push_back (*it);
    }

  // Now output given oparray to file

#ifdef VALGRIND_MAKE_READABLE
  VALGRIND_MAKE_READABLE (&op_array.type, (size_t)&op_array.function_name - (size_t)&op_array.type);
  VALGRIND_MAKE_READABLE (&op_array.uses_globals, (size_t)&op_array.static_variables - (size_t)&op_array.uses_globals);
  VALGRIND_MAKE_READABLE (&op_array.return_reference, (size_t)&op_array.filename - (size_t)&op_array.return_reference);
  VALGRIND_MAKE_READABLE (&op_array.reserved, sizeof (op_array.reserved));
#endif

  zend_op_array woa = op_array;

  data.append_ref (woa.arg_types, woa.arg_types ? woa.arg_types[0] + 1 : 0);
  data.append_rstr (woa.function_name);
  woa.refcount = NULL;
  woa.opcodes = NULL;
  woa.size = cells.size();

  // Construct the brk_cont_array
  if ( brk_cont.empty() )
    {
      woa.brk_cont_array = NULL;
      data.append_ref (woa.brk_cont_array, 0);
    }
  else
    {
      zend_brk_cont_element bca[brk_cont.size()];
      for ( size_t i = 0; i < brk_cont.size(); i++ )
        {
          bca[i].parent = brk_cont[i].parent;
          bca[i].cont = brk_cont[i].cont ? brk_cont[i].cont->order_id : -1;
          if ( brk_cont[i].brk )
              bca[i].brk = brk_cont[i].brk->order_id;
          else
              bca[i].brk = -1;
        }
      woa.brk_cont_array = bca;
      data.append_ref (woa.brk_cont_array, sizeof (zend_brk_cont_element) * brk_cont.size());
    }

  woa.static_variables = data.serialize_hashtable_of_zval (woa.static_variables, consts);
  woa.start_op = NULL;
  data.append_rstr (woa.filename);

  out.write (reinterpret_cast<char *>(&woa), sizeof (zend_op_array));

  deque<unsigned char> buf;
  size_t start = sizeof (ex_node) * cells.size();

  for ( cells_t::iterator it = cells.begin(); it != cells.end(); it++ )
    {
      Cell &cell = **it;
      ex_node &node = cell.node;

      if ( cell.next_jmp ) node.jex.alt_jump.next = reinterpret_cast<ex_node*>(cell.next_jmp->order_id);

      if ( cell.op1.op_type == IS_CONST )
          cell.node.op1.u.cnst = data.serialize_zval (&cell.op1.u.constant, consts);

      if ( cell.op2.op_type == IS_CONST )
          cell.node.op2.u.cnst = data.serialize_zval (&cell.op2.u.constant, consts);

      if ( cell.node.op2.op_type == JOP_JMP )
          cell.node.op2.u.opline_num = cell.j2_next->order_id;

      buf.insert (buf.end(), reinterpret_cast<unsigned char *>(&node), reinterpret_cast<unsigned char *>(&node + 1));
    }

  for ( cells_t::iterator it = cells.begin(); it != cells.end(); it++ )
      delete *it;

  vector<char> tmpbuf;

  tmpbuf.reserve (buf.size());
  tmpbuf.assign (buf.begin(), buf.end());

  out.write (&tmpbuf[0], buf.size());
}


