
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include "jphp.hh"

#include "shared.hh"
#include "mm.hh"

// #include "dump.hh"

#include <fcntl.h>
#include <unistd.h>
#include <utime.h>

struct shared_mem_header {
  size_t magic;
  size_t size;
  void *ptr;
  MemMan *shared_memory_heap;

  int max_cached_script_size;
  int max_cached_script_ttl;

  jphp_header *delete_list;

  size_t hash_size;
  jphp_header **hash;

  jphp_header *first;
  jphp_header *last;

  size_t elem_count;

  void insert (jphp_header *head)
  {
    size_t pos = zend_hash_func (head->shared_path, strlen (head->shared_path)) % hash_size;

    // Insert into the hash table
    if ( hash[pos] ) hash[pos]->h_prev = head;
    head->h_next = hash[pos];
    hash[pos] = head;

    // Insert into the usage order list
    head->next = first;
    if ( first )
        first->prev = head;
    else
        last = head;

    first = head;

    elem_count++;
  }

  jphp_header * find (const char *path)
  {
    size_t pos = zend_hash_func (const_cast<char *>(path), strlen (path)) % hash_size;
    jphp_header *head = hash[pos];
    while (head)
      {
        if ( strcmp (path, head->shared_path) == 0 ) break;
        head = head->h_next;
      }
    return head;
  }

  void advance (jphp_header *head)
  {
    // at first remove from list
    if ( head->prev )
        head->prev->next = head->next;
    else
        return; // already at the top of list

    if ( head->next )
        head->next->prev = head->prev;
    else
        last = head->prev;

    // and then reattach to the head of list
    first->prev = head;
    head->next = first;
    head->prev = NULL;

    first = head;
  }

  void release (jphp_header *head)
  {
    size_t pos = zend_hash_func (head->shared_path, strlen (head->shared_path)) % hash_size;

    // Remove from the hash table
    if ( head->h_next )
        head->h_next->h_prev = head->h_prev;

    if ( head->h_prev )
        head->h_prev->h_next = head->h_next;
    else
        hash[pos] = head->h_next;

    head->h_next = head->h_prev = NULL;

    // Remove from the usage list
    if ( head->prev )
        head->prev->next = head->next;
    else
        first = head->next;

    if ( head->next )
        head->next->prev = head->prev;
    else
        last = head->prev;

    // Add into delete list
    if ( delete_list )
        delete_list->prev = head;

    head->next = delete_list;
    head->prev = NULL;
    delete_list = head;

    elem_count--;
  }

  void free_memory ()
  {
    // at first retire 1/7 of active scripts, from the end of usage list

    int retire = 1 + elem_count/7;

    jphp_header *head = last;

    while (retire && head)
      {
        jphp_header *tf = head;
        head = head->prev;
        release (tf);
        retire--;
      }

    time_t now = time (NULL);

    head = delete_list;

    while (head)
      {
        if ( head->user_count == 0 || difftime (now, head->script_last_accessed) >= max_cached_script_ttl )
          {
            jphp_header *tf = head;

            if ( head->prev )
                head->prev->next = head->next;
            else
                delete_list = head->next;

            if ( head->next )
                head->next->prev = head->prev;

            head = head->next;

            shared_memory_heap->free (tf->shared_path);
            shared_memory_heap->free (tf);
          }
        else
          {
            head = head->next;
          }
      }
  }
};

namespace {
  shared_mem_header *MEM;
  SharedMem *shared_mem;
  Lock *memory_lock;
  const size_t SHARED_MEM_MAGIC = 0xc31f4a33;
  bool is_initialized = false;
  mode_t file_perm = 0400;
  mode_t dir_perm = 0700;
  mode_t lock_perm = 0600;
};

static void make_dirs (const vstring& str)
{
  mode_t m = umask(0);
  char cstr[str.size() + 1];
  strcpy (cstr, str.c_str());
  char *p = cstr;
  p++;
  while (*p)
    {
      if ( *p == '/' )
        {
          *p = 0;
          int ret = mkdir (cstr, dir_perm);
          *p = '/';
        }
      p++;
    }
  umask (m);
}

void JphpSystem::init ()
{
  if ( is_initialized ) return;  // already initialized
  is_initialized = true;

  static int v = 10;
  v++;

  if ( !INI_BOOL ("jit.enable") || !INI_BOOL ("jit.enable_file_cache") ) return;

  file_perm = strtol (INI_STR("jit.file_perm"), NULL, 8);
  dir_perm = strtol (INI_STR("jit.dir_perm"), NULL, 8);
  lock_perm = strtol (INI_STR("jit.lock_perm"), NULL, 8);

  mode_t m = umask(0);
  mkdir (INI_STR("jit.data_dir"), 01777);
  umask (m);

  if ( !INI_BOOL ("jit.enable_shared_memory_cache") ) return;

  size_t size = INI_INT("jit.shared_memory_size");

  memory_lock = new FileLock (lock_perm);

  if ( !memory_lock->init (vstring(INI_STR("jit.data_dir")) + "/shared_memory.LCK") )
    {
      cerr << "JIT-PHP Warning: unable to initialize the shared memory lock (" << strerror (errno) <<
          "), shared memory disabled!\n";
      return;
    }

  memory_lock->lock();

  shared_mem = new IPCSharedMem ();

  bool can_exit = false;
  void *base_ptr = NULL;

  while (!can_exit )
    {
      switch ( shared_mem->init (base_ptr, INI_STR("jit.shared_memory_key"), size) )
        {
          case SharedMem::error:
            delete shared_mem;
            shared_mem = NULL;
            cerr << "JIT-PHP Warning: unable to initialize the shared memory segment!\n";
            can_exit = true;
            break;
          case SharedMem::attached:
            MEM = static_cast<shared_mem_header *>(shared_mem->get_mem());

            if ( MEM->ptr != shared_mem->get_mem() && !base_ptr )
              {
                base_ptr = MEM->ptr;
                can_exit = false;
                MEM = NULL;
              }
            else if ( MEM->size != size || MEM->magic != SHARED_MEM_MAGIC || MEM->ptr != shared_mem->get_mem() )
              {
                delete shared_mem;
                shared_mem = NULL;
                MEM = NULL;
                cerr << "JIT-PHP Warning: attached shared memory parameters do not match, failed to attach!\n";
                can_exit = true;
              }
            else
              {
                can_exit = true;
              }
            break;
          case SharedMem::created:
            char *ptr = static_cast<char *>(shared_mem->get_mem());
            MEM = static_cast<shared_mem_header *>(shared_mem->get_mem());

            MEM->magic = SHARED_MEM_MAGIC;
            MEM->size = size;
            MEM->ptr = shared_mem->get_mem();
            MEM->shared_memory_heap = new (ptr + sizeof (shared_mem_header)) MemMan (
                ptr + sizeof (shared_mem_header) + sizeof (MemMan),
                size - sizeof (shared_mem_header) - sizeof (MemMan));

            MEM->hash_size = INI_INT("jit.shared_memory_hash_size");
            MEM->hash = static_cast<jphp_header **>(MEM->shared_memory_heap->calloc (
                sizeof (jphp_header *) * MEM->hash_size, 1));

            MEM->elem_count = 0;
            MEM->delete_list = NULL;
            MEM->first = NULL;
            MEM->last = NULL;

            MEM->max_cached_script_size = INI_INT("jit.shared_memory_max_cached_script_size");
            MEM->max_cached_script_ttl = INI_INT("jit.shared_memory_max_cached_script_ttl");

            can_exit = true;
            break;
        }
    }

  if ( INI_BOOL ("jit.shared_memory_destroy_at_exit") )
      shared_mem->delete_at_exit ();

  memory_lock->unlock();
}


Jphp * JphpSystem::get_running_context ()
{
  static Jphp *jphp = new Jphp ();
  return jphp;
}

jphp_header * JphpSystem::aquire_script (zend_file_handle *file_handle)
{
  struct stat fst;

  char *fn = file_handle->opened_path ? file_handle->opened_path : file_handle->filename;
  if ( !fn ) return NULL;

  jphp_header *ret = NULL;

  // 1) At first check whether the file is present in our shared memory
  // script cache. We also update the script's priority in the cache.
  // If not in the cache we also determine script's priority and if demanded
  // then allocate (actually always allocate) the memory inside the cache for
  // the script.

  // no streaming possible in other words
  if ( fstat (fileno(file_handle->handle.fp), &fst) != 0 ) return NULL; // this should not happen

  time_t t = fst.st_mtime;
//  fst.st_mtime = 0;

  if ( MEM && INI_BOOL("jit.enable_shared_memory_cache"))
    {
      memory_lock->lock ();

      ret = MEM->find (fn);

      if ( ret )
        {
          if ( ret->stmtime != fst.st_mtime )
            {
              MEM->release (ret);
              ret = NULL;
            }
          else
            {
              ret->user_count++;
              ret->usage_count++;
              ret->script_last_accessed = time (NULL);
              MEM->advance (ret);
            }
        }

      memory_lock->unlock ();
    }

//  fst.st_mtime = t;

  if ( ret )
    {
      zend_llist_add_element(&CG(open_files), file_handle);
      return ret;
    }

  // 2) Then we check whether we're dealing with already encoded script
  // we might just directly load in.

  // 3) Then we check whether this file has already been encoded and
  // the encoded file is not old, if not we'll just load it in.

  // 4) If it's just a simple PHP file and we have to encode it, encoding
  // always goes to a file under {jit.data_dir}/scripts/, from there following the
  // same full path the original opened file did.

  vstring vpath = vstring(INI_STR("jit.data_dir")) + "/scripts" + fn;

  int fd = open (vpath.c_str(), O_RDONLY);

  struct stat nst;
  if ( fd != -1 ) fstat (fd, &nst); // see if the script has changed in the mean time

  if ( fd == -1 || nst.st_mtime != fst.st_mtime ) // we have to reencode the script
    {
      if ( fd != -1 ) close (fd);

      make_dirs (vpath);

      FileLock encode_lock (vpath + ".LCK", lock_perm);
      encode_lock.lock (); // unlocked when we go out of scope

      // check whether somebody managed to encode it while we were getting the lock
      fd = open (vpath.c_str(), O_RDONLY);
      if ( fd != -1 ) fstat (fd, &nst);

      if ( fd == -1 || nst.st_mtime != fst.st_mtime )
        {
          if ( fd != -1 ) close (fd);

          vstring encode_path = vpath + ".TMP";

          mode_t m = umask (0);
          unlink (encode_path.c_str()); // make sure no old file is there for permission denied
          ofstream out (encode_path.c_str(), ios::binary|ios::trunc);
//          ofstream out (encode_path.c_str(), ios::binary|ios::trunc, file_perm); // XXX
          umask (m);

          if ( !out.is_open () )
            {
              zend_error (E_WARNING, "JIT PHP: Unable to open caching file '%s' for output (%s)", encode_path.c_str(), strerror (errno));
              zend_llist_add_element(&CG(open_files), file_handle);
              return NULL;
            }

          int ret = jphp_encode (file_handle, out, false);
          out.close();

          if ( ret == SUCCESS )
            {
              rename (encode_path.c_str(), vpath.c_str());
              fd = open (vpath.c_str(), O_RDONLY);
            }
          else
            {
              unlink (encode_path.c_str());
              return NULL;
            }

          struct utimbuf ubuf;
          ubuf.actime = fst.st_atime;
          ubuf.modtime = fst.st_mtime;
          utime (vpath.c_str(), &ubuf);

          fd = open (vpath.c_str(), O_RDONLY);
          fstat (fd, &nst);
        }
    }
  else
    {
      zend_llist_add_element(&CG(open_files), file_handle);
    }

  char *data = NULL;
  char *sfn = NULL;
  int is_shared = 0;

  if ( MEM && MEM->max_cached_script_size >= nst.st_size && INI_BOOL("jit.enable_shared_memory_cache") )
    {
      memory_lock->lock ();

      for ( size_t do_try = 0; do_try < 10; do_try++ )
        {
          data = static_cast<char *>(MEM->shared_memory_heap->alloc (nst.st_size));
          sfn = MEM->shared_memory_heap->strdup (fn);

          if ( !data || !sfn )
            {
              MEM->shared_memory_heap->free (sfn);
              MEM->shared_memory_heap->free (data);
              MEM->free_memory ();
            }
          else
            {
              is_shared = 1;
              break;
            }
        }

      memory_lock->unlock ();
    }

//  if ( !is_shared ) cerr << "Unable to get shared memory, in the amount of " << nst.st_size << ", free stands at " << MEM->shared_memory_heap->available() << endl;

  if ( !data ) data = static_cast<char *>(emalloc (nst.st_size));
  read (fd, data, nst.st_size);
  close (fd);

  ret = jphp_load (data, nst.st_size);

  if ( is_shared )
    {
      ret->is_shared = is_shared;
      ret->shared_path = sfn;
      ret->user_count = 1;
      ret->usage_count = 1;
      ret->script_last_accessed = time (NULL);
      ret->stmtime = fst.st_mtime;
      memory_lock->lock ();
      MEM->insert (ret);
      memory_lock->unlock ();
    }

  return ret;
}

void JphpSystem::release_script (jphp_header *script)
{
  if ( script->is_shared )
      script->user_count--;
  else
      efree (script); // this is now safe to do as the data is not queried anymore, op_array copy remains tho
}

