
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#ifndef __utils_hh__
#define __utils_hh__

/** Simple vector based string implementation. Shouls act somehwat as normal
  * string only is not reference counted and always maintains a \0 on the end,
  * transparently to users.*/
class vstring {
private:
  typedef vector<char> s_type;

public:
  inline vstring () : s(1, '\0') { }

  inline vstring (const vstring& c) : s (c.s) {}

  inline vstring (const char *str) : s (str, str + strlen(str) + 1) {}

  inline vstring (const char *str, size_t len) : s (str, str + len) { s.push_back ('\0'); }

  inline vstring (const string& str) : s (str.begin(), str.end()) { s.push_back ('\0'); }

  friend inline bool operator== (const vstring& l, const vstring& r)
  {
    return l.size() == r.size () && memcmp (l.c_str(), r.c_str(), l.size()) == 0;
  }

  friend vstring operator+ (const vstring& l, const vstring& r)
  {
    vstring ret(l);
    ret.s.insert (ret.end(), r.begin(), r.end());
    return ret;
  }

  friend inline bool operator< (const vstring& l, const vstring& r)
  {
    size_t min = l.size() < r.size() ? l.size() : r.size();
    int cmp = memcmp (l.c_str(), r.c_str(), min);
    if ( cmp == 0 ) return l.size() < r.size();
    return cmp < 0;
  }

  template<typename it>
  inline void assign (it s, it e) { s.assign (s, e); s.push_back ('\0'); }

  template<typename Pos, typename In>
  inline void insert (Pos p, In s, In e) { s.insert (p, s, e); }

  inline void reserve (size_t size) { s.reserve (size + 1); }

  inline size_t size () const { return s.size() - 1; }
  inline bool empty () const { return s.size() != 1; }

  inline const char * c_str() const { return &s[0]; }

  typedef s_type::iterator iterator;
  typedef s_type::const_iterator const_iterator;

  inline iterator begin() { return s.begin(); }
  inline iterator end() { return s.end() - 1; }
  inline const_iterator begin() const { return s.begin(); }
  inline const_iterator end() const { return s.end() - 1; }

private:
  s_type s;
};

struct vstring_hash_func {
  inline size_t operator() (const vstring& str) const
  { return zend_hash_func (const_cast<char *>(str.c_str()), str.size() + 1); }
};

struct vstring_ptr_hash_func {
  inline size_t operator() (const vstring* str) const
  { return zend_hash_func (const_cast<char *>(str->c_str()), str->size() + 1); }
};

template <typename T>
class php_hash
{
public:
  typedef T * Tp;
  inline php_hash (HashTable *h) : ht(h) {}

  inline void reset () { zend_hash_internal_pointer_reset (ht); }

  inline bool forw() { return zend_hash_move_forward (ht) == SUCCESS; }
  inline bool back() { return zend_hash_move_backward (ht) == SUCCESS; }

  inline bool get_data (Tp& ret)
  { return zend_hash_get_current_data (ht, reinterpret_cast<void **>(&ret)) == SUCCESS; }

  inline Tp get_data ()
  {
    Tp ret = NULL;
    get_data (ret);
    return ret;
  }

  inline vstring get_string_key ()
  {
    char *str;
    uint str_len;
    ulong index;
    zend_hash_get_current_key_ex (ht, &str, &str_len, &index, 0, NULL);
    if ( str[str_len - 1] == 0 )  // this is a hack
        return vstring (str, str_len - 1);
    else
        return vstring (str, str_len);
  }

  inline Tp each ()
  {
    Tp ret = NULL;
    if ( get_data (ret) ) forw();
    return ret;
  }

  inline bool quick_find (const char *str, size_t len, size_t hc, Tp& ret)
  { return zend_hash_quick_find(ht, const_cast<char *>(str), len, hc, reinterpret_cast<void **>(&ret)) == SUCCESS; }

  inline bool find (const char *str, size_t len, Tp& ret)
  { return zend_hash_find(ht, const_cast<char *>(str), len, reinterpret_cast<void **>(&ret)) == SUCCESS; }

  inline bool find (const char *str, Tp& ret)
  { return find (str, strlen(str) + 1, ret); }

  inline Tp find (const char *str)
  {
    Tp ret = NULL;
    find (str, ret);
    return ret;
  }

  inline Tp find (const vstring& str)
  {
    Tp ret = NULL;
    find (str.c_str(), str.size() + 1, ret);
    return ret;
  }

  inline Tp find (const char *str, size_t len)
  {
    Tp ret = NULL;
    find (str, len, ret);
    return ret;
  }

  inline bool quick_inline_find_copy (const char *key, size_t len, size_t hc, T& ret)
  {
    Bucket *p = ht->arBuckets[hc & ht->nTableMask];

    while (p != NULL) {
      if ((p->h == hc) && (p->nKeyLength == len)) {
        if (!memcmp(p->arKey, key, len)) {
          ret = *reinterpret_cast<Tp>(p->pData);
          return true;
        }
      }
      p = p->pNext;
    }
    return false;
  }

  inline bool quick_inline_find (const char *key, size_t len, size_t hc, Tp& ret)
  {
    Bucket *p = ht->arBuckets[hc & ht->nTableMask];

    while (p != NULL) {
      if ((p->h == hc) && (p->nKeyLength == len)) {
        if (!memcmp(p->arKey, key, len)) {
          ret = reinterpret_cast<Tp>(p->pData);
          return true;
        }
      }
      p = p->pNext;
    }
    return false;
  }

  inline bool insert (const char *str, size_t len, const T& toadd)
  { return zend_hash_add (ht, const_cast<char *>(str), len, reinterpret_cast<void *>(&const_cast<T&>(toadd)), sizeof(T), NULL) == SUCCESS; }

  inline bool quick_insert (const char *str, size_t len, size_t hc, const T& toadd)
  { return zend_hash_quick_add (ht, const_cast<char *>(str), len, hc, reinterpret_cast<void *>(&const_cast<T&>(toadd)), sizeof(T), NULL) == SUCCESS; }

  inline bool quick_insert (const char *str, size_t len, size_t hc, const T& toadd, Tp& ret)
  { return zend_hash_quick_add (ht, const_cast<char *>(str), len, hc, reinterpret_cast<void *>(&const_cast<T&>(toadd)), sizeof(T), (void **)&ret) == SUCCESS; }

  inline bool insert (const char *str, const T& toadd)
  { return insert (str, strlen (str) + 1, toadd); }

  inline size_t size () const
  { return zend_hash_num_elements (ht); }

private:
  mutable HashTable *ht;
};


#endif // __utils_hh__

